# Adapted from this response in Stackoverflow
# http://stackoverflow.com/a/19053800/1072990
def to_camel_case(snake_str: str) -> str:
    components = snake_str.split("_")
    # We capitalize the first letter of each component except the first one
    # with the 'capitalize' method and join them together.
    return components[0] + "".join(x.capitalize() if x else "_" for x in components[1:])


def capitalize_first(name: str) -> str:
    return name[0].upper() + name[1:]



from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Optional

from graphql import GraphQLFormattedError

from strawberry.unset import UNSET


@dataclass
class GraphQLTransportMessage:
    def as_dict(self) -> dict:
        data = asdict(self)
        if getattr(self, "payload", None) is UNSET:
            # Unset fields must have a JSON value of "undefined" not "null"
            data.pop("payload")
        return data


@dataclass
class ConnectionInitMessage(GraphQLTransportMessage):
    """
    Direction: Client -> Server
    """

    payload: Optional[Dict[str, Any]] = UNSET
    type: str = "connection_init"


@dataclass
class ConnectionAckMessage(GraphQLTransportMessage):
    """
    Direction: Server -> Client
    """

    payload: Optional[Dict[str, Any]] = UNSET
    type: str = "connection_ack"


@dataclass
class PingMessage(GraphQLTransportMessage):
    """
    Direction: bidirectional
    """

    payload: Optional[Dict[str, Any]] = UNSET
    type: str = "ping"


@dataclass
class PongMessage(GraphQLTransportMessage):
    """
    Direction: bidirectional
    """

    payload: Optional[Dict[str, Any]] = UNSET
    type: str = "pong"


@dataclass
class SubscribeMessagePayload:
    query: str
    operationName: Optional[str] = None
    variables: Optional[Dict[str, Any]] = None
    extensions: Optional[Dict[str, Any]] = None


@dataclass
class SubscribeMessage(GraphQLTransportMessage):
    """
    Direction: Client -> Server
    """

    id: str
    payload: SubscribeMessagePayload
    type: str = "subscribe"


@dataclass
class NextMessage(GraphQLTransportMessage):
    """
    Direction: Server -> Client
    """

    id: str
    payload: Dict[str, Any]  # TODO: shape like ExecutionResult
    type: str = "next"


@dataclass
class ErrorMessage(GraphQLTransportMessage):
    """
    Direction: Server -> Client
    """

    id: str
    payload: List[GraphQLFormattedError]
    type: str = "error"


@dataclass
class CompleteMessage(GraphQLTransportMessage):
    """
    Direction: bidirectional
    """

    id: str
    type: str = "complete"



import hashlib
from inspect import isawaitable
from typing import Optional

from ddtrace import tracer

from strawberry.extensions import Extension
from strawberry.extensions.tracing.utils import should_skip_tracing
from strawberry.types.execution import ExecutionContext
from strawberry.utils.cached_property import cached_property


class DatadogTracingExtension(Extension):
    def __init__(
        self,
        *,
        execution_context: Optional[ExecutionContext] = None,
    ):
        if execution_context:
            self.execution_context = execution_context

    @cached_property
    def _resource_name(self):
        query_hash = self.hash_query(self.execution_context.query)

        if self.execution_context.operation_name:
            return f"{self.execution_context.operation_name}:{query_hash}"

        return query_hash

    def hash_query(self, query: str):
        return hashlib.md5(query.encode("utf-8")).hexdigest()

    def on_request_start(self) -> None:
        self._operation_name = self.execution_context.operation_name
        span_name = (
            f"{self._operation_name}" if self._operation_name else "Anonymous Query"
        )

        self.request_span = tracer.trace(
            span_name,
            resource=self._resource_name,
            span_type="graphql",
            service="strawberry",
        )
        self.request_span.set_tag("graphql.operation_name", self._operation_name)

        operation_type = "query"

        if self.execution_context.query.strip().startswith("mutation"):
            operation_type = "mutation"
        if self.execution_context.query.strip().startswith("subscription"):
            operation_type = "subscription"

        self.request_span.set_tag("graphql.operation_type", operation_type)

    def on_request_end(self) -> None:
        self.request_span.finish()

    def on_validation_start(self):
        self.validation_span = tracer.trace("Validation", span_type="graphql")

    def on_validation_end(self):
        self.validation_span.finish()

    def on_parsing_start(self):
        self.parsing_span = tracer.trace("Parsing", span_type="graphql")

    def on_parsing_end(self):
        self.parsing_span.finish()

    async def resolve(self, _next, root, info, *args, **kwargs):
        if should_skip_tracing(_next, info):
            result = _next(root, info, *args, **kwargs)

            if isawaitable(result):  # pragma: no cover
                result = await result

            return result

        field_path = f"{info.parent_type}.{info.field_name}"

        with tracer.trace(f"Resolving: {field_path}", span_type="graphql") as span:
            span.set_tag("graphql.field_name", info.field_name)
            span.set_tag("graphql.parent_type", info.parent_type.name)
            span.set_tag("graphql.field_path", field_path)
            span.set_tag("graphql.path", ".".join(map(str, info.path.as_list())))

            result = _next(root, info, *args, **kwargs)

            if isawaitable(result):
                result = await result

            return result


class DatadogTracingExtensionSync(DatadogTracingExtension):
    def resolve(self, _next, root, info, *args, **kwargs):
        if should_skip_tracing(_next, info):
            return _next(root, info, *args, **kwargs)

        field_path = f"{info.parent_type}.{info.field_name}"

        with tracer.trace(f"Resolving: {field_path}", span_type="graphql") as span:
            span.set_tag("graphql.field_name", info.field_name)
            span.set_tag("graphql.parent_type", info.parent_type.name)
            span.set_tag("graphql.field_path", field_path)
            span.set_tag("graphql.path", ".".join(map(str, info.path.as_list())))

            return _next(root, info, *args, **kwargs)



from io import BytesIO
from typing import Dict, List, Union

from sanic.request import Request


def convert_request_to_files_dict(request: Request) -> dict:
    """
    request.files has the following format, even if only a single file is uploaded:

    {
        'textFile': [
            sanic.request.File(
                type='text/plain',
                body=b'strawberry',
                name='textFile.txt'
            )
        ]
    }

    Note that the dictionary entries are lists.
    """
    request_files: dict = request.files
    files_dict: Dict[str, Union[BytesIO, List[BytesIO]]] = {}

    for field_name, file_list in request_files.items():
        assert len(file_list) == 1
        files_dict[field_name] = BytesIO(file_list[0].body)

    return files_dict



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['strawberry',
 'strawberry.aiohttp',
 'strawberry.aiohttp.handlers',
 'strawberry.aiohttp.test',
 'strawberry.asgi',
 'strawberry.asgi.handlers',
 'strawberry.asgi.test',
 'strawberry.chalice',
 'strawberry.channels',
 'strawberry.channels.handlers',
 'strawberry.cli',
 'strawberry.cli.commands',
 'strawberry.cli.utils',
 'strawberry.codegen',
 'strawberry.codegen.plugins',
 'strawberry.django',
 'strawberry.django.test',
 'strawberry.experimental',
 'strawberry.experimental.pydantic',
 'strawberry.ext',
 'strawberry.ext.dataclasses',
 'strawberry.extensions',
 'strawberry.extensions.tracing',
 'strawberry.fastapi',
 'strawberry.fastapi.handlers',
 'strawberry.federation',
 'strawberry.file_uploads',
 'strawberry.flask',
 'strawberry.printer',
 'strawberry.sanic',
 'strawberry.schema',
 'strawberry.schema.types',
 'strawberry.subscriptions',
 'strawberry.subscriptions.protocols',
 'strawberry.subscriptions.protocols.graphql_transport_ws',
 'strawberry.subscriptions.protocols.graphql_ws',
 'strawberry.test',
 'strawberry.tools',
 'strawberry.types',
 'strawberry.types.fields',
 'strawberry.utils']

package_data = \
{'': ['*'], 'strawberry': ['static/*']}

install_requires = \
['graphql-core>=3.2.0,<3.3.0',
 'python-dateutil>=2.7.0,<3.0.0',
 'typing_extensions>=3.7.4,<5.0.0']

extras_require = \
{':python_version < "3.8"': ['backports.cached-property>=1.0.2,<2.0.0'],
 'aiohttp': ['aiohttp>=3.7.4.post0,<4.0.0'],
 'asgi': ['starlette>=0.13.6', 'python-multipart>=0.0.5,<0.0.6'],
 'chalice': ['chalice>=1.22,<2.0'],
 'channels': ['asgiref>=3.2,<4.0', 'channels>=3.0.5'],
 'cli': ['click>=7.0,<9.0', 'pygments>=2.3,<3.0'],
 'debug-server': ['starlette>=0.13.6',
                  'click>=7.0,<9.0',
                  'pygments>=2.3,<3.0',
                  'uvicorn>=0.11.6,<0.19.0',
                  'python-multipart>=0.0.5,<0.0.6'],
 'django': ['Django>=3.2', 'asgiref>=3.2,<4.0'],
 'fastapi': ['python-multipart>=0.0.5,<0.0.6', 'fastapi>=0.65.2'],
 'flask': ['flask>=1.1'],
 'opentelemetry': ['opentelemetry-api<2', 'opentelemetry-sdk<2'],
 'pydantic': ['pydantic<2'],
 'sanic': ['sanic>=20.12.2,<22.0.0']}

entry_points = \
{'console_scripts': ['strawberry = strawberry.cli:run']}

setup_kwargs = {
    'name': 'strawberry-graphql',
    'version': '0.133.5',
    'description': 'A library for creating GraphQL APIs',
    'long_description': '<img src="https://github.com/strawberry-graphql/strawberry/raw/main/.github/logo.png" width="124" height="150">\n\n# Strawberry GraphQL\n\n> Python GraphQL library based on dataclasses\n\n[![CircleCI](https://img.shields.io/circleci/token/307b40d5e152e074d34f84d30d226376a15667d5/project/github/strawberry-graphql/strawberry/main.svg?style=for-the-badge)](https://circleci.com/gh/strawberry-graphql/strawberry/tree/main)\n[![Discord](https://img.shields.io/discord/689806334337482765?label=discord&logo=discord&logoColor=white&style=for-the-badge&color=blue)](https://discord.gg/ZkRTEJQ)\n[![PyPI](https://img.shields.io/pypi/v/strawberry-graphql?logo=pypi&logoColor=white&style=for-the-badge)](https://pypi.org/project/strawberry-graphql/)\n\n## Installation ( Quick Start )\n\nThe quick start method provides a server and CLI to get going quickly. Install\nwith:\n\n```shell\npip install \'strawberry-graphql[debug-server]\'\n```\n\n## Getting Started\n\nCreate a file called `app.py` with the following code:\n\n```python\nimport strawberry\n\n\n@strawberry.type\nclass User:\n    name: str\n    age: int\n\n\n@strawberry.type\nclass Query:\n    @strawberry.field\n    def user(self) -> User:\n        return User(name="Patrick", age=100)\n\n\nschema = strawberry.Schema(query=Query)\n```\n\nThis will create a GraphQL schema defining a `User` type and a single query\nfield `user` that will return a hardcoded user.\n\nTo run the debug server run the following command:\n\n```shell\nstrawberry server app\n```\n\nOpen the debug server by clicking on the following link:\n[http://0.0.0.0:8000/graphql](http://0.0.0.0:8000/graphql)\n\nThis will open GraphiQL where you can test the API.\n\n### Type-checking\n\nStrawberry comes with a [mypy] plugin that enables statically type-checking your\nGraphQL schema. To enable it, add the following lines to your `mypy.ini`\nconfiguration:\n\n```ini\n[mypy]\nplugins = strawberry.ext.mypy_plugin\n```\n\n[mypy]: http://www.mypy-lang.org/\n\n### Django Integration\n\nA Django view is provided for adding a GraphQL endpoint to your application.\n\n1. Add the app to your `INSTALLED_APPS`.\n\n```python\nINSTALLED_APPS = [\n    ...\n    \'strawberry.django\',\n]\n```\n\n2. Add the view to your `urls.py` file.\n\n```python\nfrom strawberry.django.views import GraphQLView\nfrom .schema import schema\n\nurlpatterns = [\n    ...,\n    path(\'graphql\', GraphQLView.as_view(schema=schema)),\n]\n```\n\n## WebSockets\n\nTo support graphql Subscriptions over WebSockets you need to provide a WebSocket\nenabled server. The debug server can be made to support WebSockets with these\ncommands:\n\n```shell\npip install \'strawberry-graphql[debug-server]\'\npip install \'uvicorn[standard]\'\n```\n\n## Examples\n\n* [Various examples on how to use Strawberry](https://github.com/strawberry-graphql/examples)\n* [Full stack example using Starlette, SQLAlchemy, Typescript codegen and Next.js](https://github.com/jokull/python-ts-graphql-demo)\n* [Quart + Strawberry tutorial](https://github.com/rockyburt/Ketchup)\n\n## Contributing\n\nWe use [poetry](https://github.com/sdispater/poetry) to manage dependencies, to\nget started follow these steps:\n\n```shell\ngit clone https://github.com/strawberry-graphql/strawberry\ncd strawberry\npoetry install\npoetry run pytest\n```\n\nThis will install all the dependencies (including dev ones) and run the tests.\n\n### Pre commit\n\nWe have a configuration for\n[pre-commit](https://github.com/pre-commit/pre-commit), to add the hook run the\nfollowing command:\n\n```shell\npre-commit install\n```\n\n## Links\n\n- Project homepage: https://strawberry.rocks\n- Repository: https://github.com/strawberry-graphql/strawberry\n- Issue tracker: https://github.com/strawberry-graphql/strawberry/issues\n  - In case of sensitive bugs like security vulnerabilities, please contact\n    patrick.arminio@gmail.com directly instead of using the issue tracker. We\n    value your effort to improve the security and privacy of this project!\n\n## Licensing\n\nThe code in this project is licensed under MIT license. See [LICENSE](./LICENSE)\nfor more information.\n',
    'author': 'Patrick Arminio',
    'author_email': 'patrick.arminio@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://strawberry.rocks/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from collections import defaultdict
from copy import copy
from itertools import chain
from typing import Any, Dict, Iterable, List, Optional, Type, Union, cast

from graphql import (
    ExecutionContext as GraphQLExecutionContext,
    GraphQLField,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLUnionType,
)
from graphql.type.definition import GraphQLArgument

from strawberry.custom_scalar import ScalarDefinition, ScalarWrapper
from strawberry.enum import EnumDefinition
from strawberry.extensions import Extension
from strawberry.schema.types.concrete_type import TypeMap
from strawberry.types.types import TypeDefinition
from strawberry.union import StrawberryUnion
from strawberry.utils.cached_property import cached_property
from strawberry.utils.inspect import get_func_args

from ..printer import print_schema
from ..schema import Schema as BaseSchema
from ..schema.config import StrawberryConfig


class Schema(BaseSchema):
    def __init__(
        self,
        query: Optional[Type] = None,
        mutation: Optional[Type] = None,
        subscription: Optional[Type] = None,
        # TODO: we should update directives' type in the main schema
        directives: Iterable[Type] = (),
        types: Iterable[Type] = (),
        extensions: Iterable[Union[Type[Extension], Extension]] = (),
        execution_context_class: Optional[Type[GraphQLExecutionContext]] = None,
        config: Optional[StrawberryConfig] = None,
        scalar_overrides: Optional[
            Dict[object, Union[ScalarWrapper, ScalarDefinition]]
        ] = None,
        schema_directives: Iterable[object] = (),
        enable_federation_2: bool = False,
    ):

        query = self._get_federation_query_type(query)

        super().__init__(
            query=query,
            mutation=mutation,
            subscription=subscription,
            directives=directives,  # type: ignore
            types=types,
            extensions=extensions,
            execution_context_class=execution_context_class,
            config=config,
            scalar_overrides=scalar_overrides,
            schema_directives=schema_directives,
        )

        self._add_scalars()
        self._add_entities_to_query()

        if enable_federation_2:
            self._add_link_directives()
        else:
            self._remove_resolvable_field()

    def _get_federation_query_type(self, query: Optional[Type]) -> Type:
        """Returns a new query type that includes the _service field.

        If the query type is provided, it will be used as the base for the new
        query type. Otherwise, a new query type will be created.

        Federation needs the following two fields to be present in the query type:
        - _service: This field is used by the gateway to query for the capabilities
            of the federated service.
        - _entities: This field is used by the gateway to query for the entities
            that are part of the federated service.

        The _service field is added by default, but the _entities field is only
        added if the schema contains an entity type.
        """

        # note we don't add the _entities field here, as we need to know if the
        # schema contains an entity type first and we do that by leveraging
        # the schema converter type map, so we don't have to do that twice
        # TODO: ideally we should be able to do this without using the schema
        # converter, but for now this is the easiest way to do it
        # see `_add_entities_to_query`

        import strawberry
        from strawberry.tools.create_type import create_type
        from strawberry.tools.merge_types import merge_types

        @strawberry.type(name="_Service")
        class Service:
            sdl: str = strawberry.field(
                resolver=lambda: print_schema(self),
            )

        @strawberry.field(name="_service")
        def service() -> Service:
            return Service()

        fields = [service]

        FederationQuery = create_type(name="Query", fields=fields)

        if query is None:
            return FederationQuery

        query_type = merge_types(
            "Query",
            (
                FederationQuery,
                query,
            ),
        )

        # TODO: this should be probably done in merge_types
        if query._type_definition.extend:
            query_type._type_definition.extend = True  # type: ignore

        return query_type

    def _add_entities_to_query(self):
        entity_type = _get_entity_type(self.schema_converter.type_map)

        if not entity_type:
            return

        self._schema.type_map[entity_type.name] = entity_type
        fields = {"_entities": self._get_entities_field(entity_type)}

        # Copy the query type, update it to use the modified fields
        query_type = cast(GraphQLObjectType, self._schema.query_type)
        fields.update(query_type.fields)

        query_type = copy(query_type)
        query_type._fields = fields

        self._schema.query_type = query_type
        self._schema.type_map[query_type.name] = query_type

    def entities_resolver(self, root, info, representations):
        results = []

        for representation in representations:
            type_name = representation.pop("__typename")
            type_ = self.schema_converter.type_map[type_name]

            definition = cast(TypeDefinition, type_.definition)
            resolve_reference = definition.origin.resolve_reference

            func_args = get_func_args(resolve_reference)
            kwargs = representation

            if "info" in func_args:
                kwargs["info"] = info

            results.append(resolve_reference(**kwargs))

        return results

    def _add_scalars(self):
        self.Any = GraphQLScalarType("_Any")

        self._schema.type_map["_Any"] = self.Any

    def _remove_resolvable_field(self) -> None:
        # this might be removed when we remove support for federation 1
        # or when we improve how we print the directives
        from ..unset import UNSET
        from .schema_directives import Key

        for directive in self.schema_directives_in_use:
            if isinstance(directive, Key):
                directive.resolvable = UNSET

    @cached_property
    def schema_directives_in_use(self) -> List[object]:
        all_graphql_types = self._schema.type_map.values()

        directives = []

        for type_ in all_graphql_types:
            strawberry_definition = type_.extensions.get("strawberry-definition")

            if not strawberry_definition:
                continue

            directives.extend(strawberry_definition.directives)

            fields = getattr(strawberry_definition, "fields", [])
            values = getattr(strawberry_definition, "values", [])

            for field in chain(fields, values):
                directives.extend(field.directives)

        return directives

    def _add_link_directives(self):
        from .schema_directives import FederationDirective, Link

        directive_by_url = defaultdict(set)

        for directive in self.schema_directives_in_use:
            if isinstance(directive, FederationDirective):
                directive_by_url[directive.imported_from.url].add(
                    f"@{directive.imported_from.name}"
                )

        link_directives = tuple(
            Link(
                url=url,
                import_=list(sorted(directives)),
            )
            for url, directives in directive_by_url.items()
        )

        self.schema_directives = tuple(self.schema_directives) + link_directives

    def _get_entities_field(self, entity_type: GraphQLUnionType) -> GraphQLField:
        return GraphQLField(
            GraphQLNonNull(GraphQLList(entity_type)),
            args={
                "representations": GraphQLArgument(
                    GraphQLNonNull(GraphQLList(GraphQLNonNull(self.Any)))
                )
            },
            resolve=self.entities_resolver,
        )


def _get_entity_type(type_map: TypeMap):
    # https://www.apollographql.com/docs/apollo-server/federation/federation-spec/#resolve-requests-for-entities

    # To implement the _Entity union, each type annotated with @key
    # should be added to the _Entity union.

    federation_key_types = [
        type.implementation
        for type in type_map.values()
        if _has_federation_keys(type.definition)
        # TODO: check this
        and not isinstance(type.implementation, GraphQLInterfaceType)
    ]

    # If no types are annotated with the key directive, then the _Entity
    # union and Query._entities field should be removed from the schema.
    if not federation_key_types:
        return None

    entity_type = GraphQLUnionType("_Entity", federation_key_types)  # type: ignore

    def _resolve_type(self, value, _type):
        return self._type_definition.name

    entity_type.resolve_type = _resolve_type

    return entity_type


def _is_key(directive: Any) -> bool:
    from .schema_directives import Key

    return isinstance(directive, Key)


def _has_federation_keys(
    definition: Union[TypeDefinition, ScalarDefinition, EnumDefinition, StrawberryUnion]
) -> bool:
    if isinstance(definition, TypeDefinition):
        return any(_is_key(directive) for directive in definition.directives or [])

    return False



from __future__ import annotations

from typing import Any, Dict, TypeVar

from pydantic import BaseModel
from typing_extensions import Protocol

from strawberry.types.types import TypeDefinition


PydanticModel = TypeVar("PydanticModel", bound=BaseModel)


class StrawberryTypeFromPydantic(Protocol[PydanticModel]):
    """This class does not exist in runtime.
    It only makes the methods below visible for IDEs"""

    def __init__(self, **kwargs):
        ...

    @staticmethod
    def from_pydantic(
        instance: PydanticModel, extra: Dict[str, Any] = None
    ) -> StrawberryTypeFromPydantic[PydanticModel]:
        ...

    def to_pydantic(self, **kwargs) -> PydanticModel:
        ...

    @property
    def _type_definition(self) -> TypeDefinition:
        ...

    @property
    def _pydantic_type(self) -> PydanticModel:
        ...



from typing import Iterable, Optional

from strawberry.arguments import StrawberryArgumentAnnotation


def argument(
    description: Optional[str] = None,
    name: Optional[str] = None,
    deprecation_reason: Optional[str] = None,
    directives: Iterable[object] = (),
    inaccessible: bool = False,
    tags: Optional[Iterable[str]] = (),
) -> StrawberryArgumentAnnotation:
    from strawberry.federation.schema_directives import Inaccessible, Tag

    directives = list(directives)

    if inaccessible:
        directives.append(Inaccessible())

    if tags:
        directives.extend(Tag(name=tag) for tag in tags)

    return StrawberryArgumentAnnotation(
        description=description,
        name=name,
        deprecation_reason=deprecation_reason,
        directives=directives,
    )
